/*
** Language Query Example Program.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctpublic.h>
#include <bkpublic.h>


#define TABLE_NAME "blktest_table"
/********************************
create table blktest_table (
	 date		int
	,item_cd	char(20)
	,iss_cd		char(20)
	,data		float
	,updt_id	char(8)
	,updt_date	int
	,updt_time	int
)
*********************************/

#ifndef LOCALE
// #  define LOCALE	"CP932"
#endif

typedef struct {
	int		date;
	char	item_cd[20+1];
	char	iss_cd[20+1];
	double	data;
	char	updt_id[8+1];
	int		updt_date;
	int		updt_time;
} Blk_Data;

Blk_Data  blk_data_buf[] = {
	{20051230, "item000", "iss000", 1.1415, "blktest", 20071110, 1200},
//	{20051231, "item001", "iss001",-9.3702, "blktest", 20071111, 1201},
};

#define blk_data_rowcnt (sizeof(blk_data_buf) / sizeof(Blk_Data))

#define MAXCOLUMNS 100
#define ERR_CH stderr
#define OUT_CH stdout

/*
** Define a macro that exits if a function return code indicates
** failure.
*/
#define EXIT_ON_FAIL(context, ret, str) \
	if (ret != CS_SUCCEED) \
	{ \
		fprintf(ERR_CH, "Fatal error: %s\n", str); \
		if (context != (CS_CONTEXT *) NULL) \
		{ \
			(CS_VOID) ct_exit(context, CS_FORCE_EXIT); \
			(CS_VOID) cs_ctx_drop(context); \
		} \
		exit(-1); \
	}
/*
** Callback routines for library errors and server messages.
*/
CS_RETCODE csmsg_callback();
CS_RETCODE clientmsg_callback();
CS_RETCODE servermsg_callback();

/*
** Main entry point for the program.
*/
int main(argc, argv)
int argc;
char **argv;
{
	CS_CONTEXT *context;		/* Context structure */
#ifdef LOCALE
	CS_LOCALE *locale;		/* Locale structure */
#endif /*LOCALE*/
	CS_CONNECTION *connection;	/* Connection structure. */
	CS_BLKDESC	*blkdesc;		/* Bulk structure. */
	CS_BOOL 	bool;
	int n, block, row, col;
	CS_DATAFMT columns[MAXCOLUMNS];
	CS_INT datalength[MAXCOLUMNS];
	//CS_SMALLINT indicator[MAXCOLUMNS];
	CS_INT count;
	CS_RETCODE ret;
	char	*server_name;
	char	*user_name;
	char	*password;
	int		blockcount = 0;
	int		batchsize = 0;
	
	/*
	** Step 1: Initialize the application.
	*/

	if (argc < 4 || argc > 6) {
		fprintf(stderr, "usage: %s server user pass [block# [batch#]]\n", argv[0]);
		exit(1);
	}
	server_name = argv[1];
	user_name = argv[2];
	password = argv[3];
	if (argc >= 5) {
		blockcount = atoi(argv[4]);
	}
	if (argc >= 6) {
		batchsize = atoi(argv[5]);
	}

	/*
	** First allocate a context structure.
	*/
	context = (CS_CONTEXT *) NULL;
	ret = cs_ctx_alloc(CS_VERSION_100, &context);
	EXIT_ON_FAIL(context, ret, "cs_ctx_alloc failed");

#ifdef LOCALE
	/*
	** Allocate locale structure.
	*/
	ret = cs_loc_alloc(context, &locale);
	EXIT_ON_FAIL(context, ret, "cs_loc_alloc failed");

	/*
	** Set locale.
	*/
	ret = cs_locale(context, CS_SET, locale, CS_LC_ALL, LOCALE, CS_NULLTERM, NULL);
	EXIT_ON_FAIL(context, ret, "cs_locale failed");
#endif /*LOCALE*/

	/*
	** Initialize Client-Library.
	*/
	ret = ct_init(context, CS_VERSION_100);
	EXIT_ON_FAIL(context, ret, "ct_init failed");
	
	/*
	** Step 2: Set up the error handling. Install
	** callback handlers for:
	** - CS-Library errors
	** - Client-Library errors
	** - Server messages.
	*/

	/*
	** Install a callback function to handle CS-Library
	** errors.
	*/
	ret = cs_config(context, CS_SET, CS_MESSAGE_CB,
						(CS_VOID *)csmsg_callback, CS_UNUSED, NULL);
	EXIT_ON_FAIL(context, ret, "cs_config(CS_MESSAGE_CB) failed");
	
	/*
	** Install a callback function to handle Client-Library
	** errors.
	**
	** The client message callback receives error or
	** informational messages discovered by
	** Client-Library.
	*/
	ret = ct_callback(context, NULL, CS_SET, CS_CLIENTMSG_CB,
						(CS_VOID *) clientmsg_callback);
	EXIT_ON_FAIL(context,ret, "ct_callback for client messages failed");
	/*
	** The server message callback receives server messages
	** sent by the server. These are error or inforamational
	** messages.
	*/
	ret = ct_callback(context, NULL, CS_SET, CS_SERVERMSG_CB,
						(CS_VOID *) servermsg_callback);
	EXIT_ON_FAIL(context, ret, "ct_callback for server messages failed");
	
	/*
	** Step 3: Connect to the server. We must:
	** - Allocate a connection structure.
	** - Set user name and password.
	** - Create the connection.
	*/

	/*
	** First, allocate a connection structure.
	*/
	ret = ct_con_alloc(context, &connection);
	EXIT_ON_FAIL(context, ret, "ct_con_alloc() failed");

	/*
	** These two calls set the user credentials (username and
	** password) for opening the connection.
	*/
	ret = ct_con_props(connection, CS_SET, CS_USERNAME,	user_name, CS_NULLTERM, NULL);
	EXIT_ON_FAIL(context, ret, "Could not set user name");

	ret = ct_con_props(connection, CS_SET, CS_PASSWORD,	password, CS_NULLTERM, NULL);
	EXIT_ON_FAIL(context, ret, "Could not set password");

	bool = CS_TRUE;
	ret = ct_con_props(connection, CS_SET, CS_BULK_LOGIN,(CS_VOID *)&bool, CS_UNUSED, NULL);
	EXIT_ON_FAIL(context, ret, "Could not set password");
	
	/*
	** Create the connection.
	*/
	ret = ct_connect(connection, (CS_CHAR *) server_name, CS_NULLTERM);
	EXIT_ON_FAIL(context, ret, "Could not connect!");

	/*
	** Step 4: 
	** - Allocate a CS_BLKDESC structure
	*/

	/*
	** Allocate a blkdesc structure.
	*/
	ret = blk_alloc(connection, BLK_VERSION_100, &blkdesc);
	EXIT_ON_FAIL(context, ret, "blk_alloc() failed");

	/*
	** Initialize blk.
	*/
	ret = blk_init(blkdesc, CS_BLK_IN, TABLE_NAME, (CS_INT)strlen(TABLE_NAME));
	EXIT_ON_FAIL(context, ret, "blk_init() failed");

	/*
	**
	*/
	n = 0;
	for(block=0 ; block < blockcount ; block++) {
		for(row=0 ; row < blk_data_rowcnt ; row++) {
			++n;
			col = -1;

			++col;
			columns[col].locale = 0;
			columns[col].count = 1;
			columns[col].datatype = CS_INT_TYPE;
			columns[col].maxlength = sizeof(blk_data_buf[row].date);
			datalength[col] = CS_UNUSED;
			ret = blk_bind(blkdesc, col+1, &columns[col], (CS_VOID *) &blk_data_buf[row].date, 
							&datalength[col], NULL);
			EXIT_ON_FAIL(context, ret, "blk_bind() failed");

			++col;
			columns[col].locale = 0;
			columns[col].count = 1;
			columns[col].datatype = CS_CHAR_TYPE;
			columns[col].maxlength = sizeof(blk_data_buf[row].item_cd)-1;
			datalength[col] = (CS_INT)strlen(blk_data_buf[row].item_cd);
			ret = blk_bind(blkdesc, col+1, &columns[col], (CS_VOID *) &blk_data_buf[row].item_cd[0], 
							&datalength[col], NULL);
			EXIT_ON_FAIL(context, ret, "blk_bind() failed");

			++col;
			columns[col].locale = 0;
			columns[col].count = 1;
			columns[col].datatype = CS_CHAR_TYPE;
			columns[col].maxlength = sizeof(blk_data_buf[row].iss_cd)-1;
			datalength[col] = (CS_INT)strlen(blk_data_buf[row].iss_cd);
			ret = blk_bind(blkdesc, col+1, &columns[col], (CS_VOID *) &blk_data_buf[row].iss_cd[0], 
							&datalength[col], NULL);
			EXIT_ON_FAIL(context, ret, "blk_bind() failed");

			++col;
			columns[col].locale = 0;
			columns[col].count = 1;
			columns[col].datatype = CS_FLOAT_TYPE;
			columns[col].maxlength = sizeof(blk_data_buf[row].data);
			datalength[col] = CS_UNUSED;
			ret = blk_bind(blkdesc, col+1, &columns[col], (CS_VOID *) &blk_data_buf[row].data, 
							&datalength[col], NULL);
			EXIT_ON_FAIL(context, ret, "blk_bind() failed");

			++col;
			columns[col].locale = 0;
			columns[col].count = 1;
			columns[col].datatype = CS_CHAR_TYPE;
			columns[col].maxlength = sizeof(blk_data_buf[row].updt_id)-1;
			datalength[col] = (CS_INT)strlen(blk_data_buf[row].updt_id);
			ret = blk_bind(blkdesc, col+1, &columns[col], (CS_VOID *) &blk_data_buf[row].updt_id[0], 
							&datalength[col], NULL);
			EXIT_ON_FAIL(context, ret, "blk_bind() failed");

			++col;
			columns[col].locale = 0;
			columns[col].count = 1;
			columns[col].datatype = CS_INT_TYPE;
			columns[col].maxlength = sizeof(blk_data_buf[row].updt_date);
			datalength[col] = CS_UNUSED;
			ret = blk_bind(blkdesc, col+1, &columns[col], (CS_VOID *) &blk_data_buf[row].updt_date, 
							&datalength[col], NULL);
			EXIT_ON_FAIL(context, ret, "blk_bind() failed");

			++col;
			columns[col].locale = 0;
			columns[col].count = 1;
			columns[col].datatype = CS_INT_TYPE;
			columns[col].maxlength = sizeof(blk_data_buf[row].updt_time);
			datalength[col] = CS_UNUSED;
			ret = blk_bind(blkdesc, col+1, &columns[col], (CS_VOID *) &blk_data_buf[row].updt_time, 
							&datalength[col], NULL);
			EXIT_ON_FAIL(context, ret, "blk_bind() failed");


			ret = blk_rowxfer(blkdesc);
			EXIT_ON_FAIL(context, ret, "blk_rowxfer() failed");

			if (batchsize != 0 && n%batchsize == 0) {
				ret = blk_done(blkdesc, CS_BLK_BATCH, &count);
				EXIT_ON_FAIL(context, ret, "blk_done() failed");
			}

			blk_data_buf[row].data += 1.0;
		}
	}
	
	/*
	** Send these rows over as a batch.
	*/
	ret = blk_done(blkdesc, CS_BLK_ALL, &count);
	EXIT_ON_FAIL(context, ret, "blk_done() failed");

	/*
	** drop blkdesc
	*/
	ret = blk_drop(blkdesc);
	EXIT_ON_FAIL(context, ret, "blk_drop() failed");
	
	/*
	** Close the connection and drop its control structure.
	*/
	ret = ct_close(connection, CS_UNUSED);
	EXIT_ON_FAIL(context, ret, "ct_close failed");
	ret = ct_con_drop(connection);
	EXIT_ON_FAIL(context, ret, "ct_con_drop failed");
	
	/*
	** ct_exit tells Client-Library that we are done.
	*/
	ret = ct_exit(context, CS_UNUSED);
	EXIT_ON_FAIL(context, ret, "ct_exit failed");
	
	/*
	** Drop the context structure.
	*/
	ret = cs_ctx_drop(context);
	EXIT_ON_FAIL(context, ret, "cs_ctx_drop failed");
	
	/*
	** Normal exit to the operating system.
	*/
	return 0;
}

/*
** Handler for server messages. Client-Library will call this
** routine when it receives a message from the server.
*/
CS_RETCODE servermsg_callback(cp, chp, msgp)
CS_CONTEXT *cp;
CS_CONNECTION *chp;
CS_SERVERMSG *msgp;
{
	/*
	** Print the message info.
	*/
	fprintf(ERR_CH, "Server message:\n\t");
	fprintf(ERR_CH,	"number(%ld) severity(%ld) state(%ld) line(%ld)\n",
			(long) msgp->msgnumber, (long) msgp->severity,
			(long) msgp->state, (long) msgp->line);
	/*
	** Print the server name if one was supplied.
	*/
	if (msgp->svrnlen > 0)
		fprintf(ERR_CH, "\tServer name: %s\n", msgp->svrname);
	/*
	** Print the procedure name if one was supplied.
	*/
	if (msgp->proclen > 0)
		fprintf(ERR_CH, "\tProcedure name: %s\n", msgp->proc);
	/*
	** Print the null terminated message.
	*/
	fprintf(ERR_CH, "\t%s\n", msgp->text);
	/*
	** Server message callbacks must return CS_SUCCEED.
	*/
	return(CS_SUCCEED);
}

/*
** Client-Library error handler. This function will be invoked
** when a Client-Library has detected an error. Before Client-
** Library routines return CS_FAIL, this handler will be called
** with additional error information.
*/
CS_RETCODE clientmsg_callback(context, conn, emsgp)
CS_CONTEXT *context;
CS_CONNECTION *conn;
CS_CLIENTMSG *emsgp;
{
	/*
	** Error number:
	** Print the error's severity, number, origin, and layer.
	** These four numbers uniquely identify the error.
	*/
	fprintf(ERR_CH,	"Client Library error:\n\t");
	fprintf(ERR_CH,	"severity(%ld) number(%ld) origin(%ld) layer(%ld)\n",
	(long) CS_SEVERITY(emsgp->severity),
	(long) CS_NUMBER(emsgp->msgnumber),
	(long) CS_ORIGIN(emsgp->msgnumber),
	(long) CS_LAYER(emsgp->msgnumber));
	/*
	** Error text:
	** Print the error text.
	*/
	fprintf(ERR_CH, "\t%s\n", emsgp->msgstring);
	/*
	** Operating system error information:
	** Some errors, such as network errors, may have
	** an operating system error associated with them.
	** If there was an operating system error,
	** this code prints the error message text.
	*/
	if (emsgp->osstringlen > 0)
	{
		fprintf(ERR_CH, "Operating system error number(%ld):\n", (long) emsgp->osnumber);
		fprintf(ERR_CH, "\t%s\n", emsgp->osstring);
	}
	/*
	** If we return CS_FAIL, Client-Library marks the connection
	** as dead. This means that it cannot be used anymore.
	** If we return CS_SUCCEED, the connection remains alive
	** if it was not already dead.
	*/
	return (CS_SUCCEED);
}

/*
** CS-Library error handler. This function will be invoked
** when CS-Library has detected an error.
*/
CS_RETCODE csmsg_callback(context, emsgp)
CS_CONTEXT *context;
CS_CLIENTMSG *emsgp;
{
	/*
	** Print the error number and message.
	*/
	fprintf(ERR_CH,	"CS-Library error:\n");
	fprintf(ERR_CH,	"\tseverity(%ld) layer(%ld) origin(%ld) number(%ld)",
	(long) CS_SEVERITY(emsgp->msgnumber),
	(long) CS_LAYER(emsgp->msgnumber),
	(long) CS_ORIGIN(emsgp->msgnumber),
	(long) CS_NUMBER(emsgp->msgnumber));
	fprintf(ERR_CH, "\t%s\n", emsgp->msgstring);
	/*
	** Print any operating system error information.
	*/
	if(emsgp->osstringlen > 0)
	{
		fprintf(ERR_CH, "Operating System Error: %s\n",	emsgp->osstring);
	}
	return (CS_SUCCEED);
}
